------------------------------------------------------------------
-- drop objects if they exist
------------------------------------------------------------------
drop table ero_deliveries;
drop table ero_returns;
drop table ero_subscriptions;
drop table ero_clients;
drop table err$_ero_deliveries;
drop table err$_ero_returns;
drop table err$_ero_subscriptions;
drop table err$_ero_clients;
drop type clients_ntt;
drop type subscriptions_ntt;
drop type deliveries_ntt;
drop type returns_ntt;
drop type toplevel_pk_ntt;
drop type client_ot;
drop type subscription_ot;
drop type delivery_ot;
drop type return_ot;
drop type toplevel_pk_ot;


------------------------------------------------------------------
-- create tables, PK's and UK's
------------------------------------------------------------------
create table ero_clients
  (id      number(10,0)   not null
  ,name    varchar2(30)   not null
  );

create unique index ero_clients_pk on ero_clients (id);
alter table ero_clients add (constraint ero_clients_pk primary key (id) using index);

create table ero_subscriptions
  (id          number(10,0)   not null
  ,client_id   number(10,0)   not null
  ,description varchar2(30)   not null
  );

create unique index ero_subscriptions_pk on ero_subscriptions (id);
alter table ero_subscriptions add constraint ero_subscriptions_pk   primary key (id) using index;
alter table ero_subscriptions add constraint ero_subscr_clients_fk1 foreign key (client_id) references ero_clients (id);

create table ero_deliveries
  (id                number(10,0)   not null
  ,subscription_id   number(10,0)   not null
  ,description       varchar2(30)   not null
  );

create unique index ero_deliveries_pk on ero_deliveries (id);
alter table ero_deliveries add constraint ero_deliveries_pk         primary key (id) using index;
alter table ero_deliveries add constraint ero_deliveries_subscr_fk1 foreign key (subscription_id) references ero_subscriptions (id);

create table ero_returns
  (id                number(10,0)   not null
  ,subscription_id   number(10,0)   not null
  ,description       varchar2(30)   not null
  );

create unique index ero_returns_pk on ero_returns (id);
alter table ero_returns add constraint ero_returns_pk         primary key (id) using index;
alter table ero_returns add constraint ero_returns_subscr_fk1 foreign key (subscription_id) references ero_subscriptions (id);

------------------------------------------------------------------
-- Create a check constraint to generate an error in the bulk DML
------------------------------------------------------------------
alter table ero_returns
add (constraint ero_returns_ck1
     check (-- description must be at least 5 chracters
            length(description) >= 5
           )
    );

------------------------------------------------------------------
-- Create log tables
------------------------------------------------------------------
begin
  dbms_errlog.create_error_log(dml_table_name => 'ero_clients');
end;
/

begin
  dbms_errlog.create_error_log(dml_table_name => 'ero_subscriptions');
end;
/

begin
  dbms_errlog.create_error_log(dml_table_name => 'ero_deliveries');
end;
/

begin
  dbms_errlog.create_error_log(dml_table_name => 'ero_returns');
end;
/

------------------------------------------------------------------
-- Create nested tables to hold the data to be inserted
------------------------------------------------------------------
create or replace type client_ot force as object
 (id      number(10,0)
 ,name    varchar2(30)
 )
/

create or replace type subscription_ot force as object
 (id          number(10,0)
 ,client_id   number(10,0)
 ,description varchar2(30)
 )
/

create or replace type delivery_ot force as object
 (id                number(10,0)
 ,subscription_id   number(10,0)
 ,description       varchar2(30)
 )
/

create or replace type return_ot force as object
 (id                number(10,0)
 ,subscription_id   number(10,0)
 ,description       varchar2(30)
 )
/

create or replace type clients_ntt force as table of client_ot
/
create or replace type subscriptions_ntt force as table of subscription_ot
/
create or replace type deliveries_ntt force as table of delivery_ot
/
create or replace type returns_ntt force as table of return_ot
/

------------------------------------------------------------------
-- Create a nested table to hold top-level PK's
------------------------------------------------------------------
create or replace type toplevel_pk_ot force as object
 (id      number(10,0)
 )
/
create or replace type toplevel_pk_ntt force as table of toplevel_pk_ot
/


------------------------------------------------------------------
-- Execute the Multi-DML code
------------------------------------------------------------------
declare
  subtype st_identifier   is varchar2(30);

  c_tbl_clients         constant   st_identifier    := 'ero_clients'      ;
  c_tbl_subscriptions   constant   st_identifier    := 'ero_subscriptions';
  c_tbl_deliveries      constant   st_identifier    := 'ero_deliveries'   ;
  c_tbl_returns         constant   st_identifier    := 'ero_returns'      ;

  l_clients_nt         clients_ntt         := clients_ntt       ();
  l_subscriptions_nt   subscriptions_ntt   := subscriptions_ntt ();
  l_deliveries_nt      deliveries_ntt      := deliveries_ntt    ();
  l_returns_nt         returns_ntt         := returns_ntt       ();

  l_client_id          ero_clients.id%type        := 0;
  l_subscription_id    ero_subscriptions.id%type  := 0;
  l_delivery_id        ero_deliveries.id%type     := 0;
  l_return_id          ero_returns.id%type        := 0;

  l_transaction_guid   varchar2(32);


  -- Print a line
  procedure pl
    (pi_line  in  varchar2  default ''
    )
    is
    begin   -- pl
      dbms_output.put_line (pi_line);
    end pl;


  -- Add a client to its nested table
  procedure add_client
    (pi_name  in  ero_clients.name%type
    )
    is
      l_client_rec  ero_clients%rowtype;
    begin   -- add_client
      -- increase id
      l_client_id := l_client_id + 1;

      -- Add the client to the nested table
      l_clients_nt.extend;
      l_clients_nt (l_clients_nt.last) := client_ot (id   => l_client_id
                                                    ,name => pi_name    
                                                    );

    end add_client;


  -- Add a subscription to its nested table
  procedure add_subscription
    (pi_description  in  ero_subscriptions.description%type
    )
    is
      l_subscription_rec  ero_subscriptions%rowtype;
    begin   -- add_subscription
      -- increase id
      l_subscription_id := l_subscription_id + 1;

      -- Add the subscription to the nested table
      l_subscriptions_nt.extend;
      l_subscriptions_nt (l_subscriptions_nt.last) := subscription_ot (id          => l_subscription_id
                                                                      ,client_id   => l_client_id      
                                                                      ,description => pi_description   
                                                                      );
    end add_subscription;


  -- Add a delivery to its nested table
  procedure add_delivery
    (pi_description      in  ero_deliveries.description%type
    )
    is
      l_delivery_rec  ero_deliveries%rowtype;
    begin   -- add_delivery
      -- increase id
      l_delivery_id := l_delivery_id + 1;

      -- Add the delivery to the nested table
      l_deliveries_nt.extend;
      l_deliveries_nt (l_deliveries_nt.last) := delivery_ot (id              => l_delivery_id    
                                                            ,subscription_id => l_subscription_id
                                                            ,description     => pi_description   
                                                            );
    end add_delivery;


  -- Add a return to its nested table
  procedure add_return
    (pi_description      in  ero_returns.description%type
    )
    is
      l_return_rec  ero_returns%rowtype;
    begin   -- add_return
      -- increase id
      l_return_id := l_return_id + 1;

      -- Add the return to the nested table
      l_returns_nt.extend;
      l_returns_nt (l_returns_nt.last) := return_ot (id              => l_return_id      
                                                    ,subscription_id => l_subscription_id
                                                    ,description     => pi_description   
                                                    );
    end add_return;


  procedure handle_bulk_errors
    (pi_table_name  in  st_identifier
    )
    is
      toplevel_pk_nt    toplevel_pk_ntt   := toplevel_pk_ntt();
      
      
      procedure get_toplevel_pks
        (pio_pk_nt   in out nocopy  toplevel_pk_ntt
        )
        is
        begin   -- get_toplevel_pks
          case pi_table_name
            when c_tbl_clients       then select toplevel_pk_ot (id => err.id)
                                          bulk collect
                                          into   pio_pk_nt
                                          from   err$_ero_clients   err
                                          ;
            when c_tbl_subscriptions then select toplevel_pk_ot (id => err.client_id)
                                          bulk collect
                                          into   pio_pk_nt
                                          from   err$_ero_subscriptions   err
                                          ;
            when c_tbl_deliveries    then select toplevel_pk_ot (id => sub.client_id)
                                          bulk collect
                                          into   pio_pk_nt
                                          from   ero_subscriptions        sub
                                          join   err$_ero_deliveries      err
                                            on   err.subscription_id = sub.id
                                          ;
            when c_tbl_returns       then select toplevel_pk_ot (id => sub.client_id)
                                          bulk collect
                                          into   pio_pk_nt
                                          from   ero_subscriptions        sub
                                          join   err$_ero_returns         err
                                            on   err.subscription_id = sub.id
                                          ;
          end case;
        exception
          when no_data_found
          then
            null;
        end get_toplevel_pks;
      
    begin   -- handle_bulk_errors
      
      -- Get the PK's of all Clients (top of the hierarchie) for which errors occured (if any!!)
      get_toplevel_pks (pio_pk_nt => toplevel_pk_nt);
      
      if toplevel_pk_nt.count > 0
      then
        pl ('ERROR occured while processing table: '||pi_table_name);
        pl ('=====================================================================');
        
        -- only perform folowing DML if errors actually occured
        delete
        from   ero_returns
        where  subscription_id in
               (select sub.id
                from   table (toplevel_pk_nt)   top
                join   ero_subscriptions        sub
                  on   sub.client_id   = top.id
               )
        ;
        
        pl ('ero_returns       records removed: '||to_char(sql%rowcount));

        delete
        from   ero_deliveries
        where  subscription_id in
               (select sub.id
                from   table (toplevel_pk_nt)   top
                join   ero_subscriptions        sub
                  on   sub.client_id   = top.id
               )
        ;
        
        pl ('ero_deliveries    records removed: '||to_char(sql%rowcount));
        
        delete
        from   ero_subscriptions
        where  client_id in
               (select top.id
                from   table (toplevel_pk_nt)   top
               )
        ;
        
        pl ('ero_subscriptions records removed: '||to_char(sql%rowcount));
        
        delete
        from   ero_clients
        where  id in
               (select top.id
                from   table (toplevel_pk_nt)   top
               )
        ;

        pl ('ero_clients       records removed: '||to_char(sql%rowcount));
        
        -- Don't forget to remove records from collections that may not have been
        -- processed yet
        begin
          select return_ot
                   (id              => ret.id              
                   ,subscription_id => ret.subscription_id 
                   ,description     => ret.description     
                   )
          bulk collect
          into   l_returns_nt
          from   table (l_returns_nt)                ret
          join   table (l_subscriptions_nt)          sub
            on   sub.id   = ret.subscription_id
          left outer join   table (toplevel_pk_nt)   top
            on   top.id   = sub.client_id
          where  top.id   is null
          ;
        exception
          when no_data_found
          then
            l_returns_nt.delete;
        end;
        
        begin
          select delivery_ot
                   (id              => del.id              
                   ,subscription_id => del.subscription_id 
                   ,description     => del.description     
                   )
          bulk collect
          into   l_deliveries_nt
          from   table (l_deliveries_nt)             del
          join   table (l_subscriptions_nt)          sub
            on   sub.id   = del.subscription_id
          left outer join   table (toplevel_pk_nt)   top
            on   top.id   = sub.client_id
          where  top.id   is null
          ;
        exception
          when no_data_found
          then
            l_deliveries_nt.delete;
        end;
        
        begin
          select subscription_ot
                   (id          => sub.id         
                   ,client_id   => sub.client_id  
                   ,description => sub.description
                   )
          bulk collect
          into   l_subscriptions_nt
          from   table (l_subscriptions_nt)          sub
          left outer join   table (toplevel_pk_nt)   top
            on   top.id   = sub.client_id
          where  top.id   is null
          ;
        exception
          when no_data_found
          then
            l_subscriptions_nt.delete;
        end;
        
        begin
          select client_ot
                   (id   => cli.id  
                   ,name => cli.name
                   )
          bulk collect
          into   l_clients_nt
          from   table (l_clients_nt)                cli
          left outer join   table (toplevel_pk_nt)   top
            on   top.id   = cli.id
          where  top.id   is null
          ;
        exception
          when no_data_found
          then
            l_subscriptions_nt.delete;
        end;

      end if;
    end handle_bulk_errors;

begin
  -- prepare data for 3 clients
  -- one returns-record for client 2 will raise an exception
  add_client         (pi_name         => 'My very first client');
    add_subscription (pi_description  => 'Fruit');
      add_delivery   (pi_description  => 'Appels');
      add_delivery   (pi_description  => 'Pears');
      add_return     (pi_description  => 'Grapes');

    add_subscription (pi_description  => 'Beverages');
      add_delivery   (pi_description  => 'Beer');
      add_delivery   (pi_description  => 'Water');
      add_return     (pi_description  => 'Water');

  add_client         (pi_name         => 'Problematic Client');
    add_subscription (pi_description  => 'Fruit');
      add_delivery   (pi_description  => 'Strawberries');
      add_delivery   (pi_description  => 'Cherries');
      add_return     (pi_description  => 'Strawberries');

    add_subscription (pi_description  => 'Beverages');
      add_delivery   (pi_description  => 'whiskey');
      add_delivery   (pi_description  => 'Lemonade');
      add_return     (pi_description  => 'Lemonade');
      add_return     (pi_description  => '&Returned_Item'); -- If this variable less than 5 characters => exception

  add_client         (pi_name         => 'Latest Client');
    add_subscription (pi_description  => 'Fruit');
      add_delivery   (pi_description  => 'Oranges');
      add_delivery   (pi_description  => 'Bananas');
      add_return     (pi_description  => 'Bananas');

    add_subscription (pi_description  => 'Beverages');
      add_delivery   (pi_description  => 'Vodka');
      add_delivery   (pi_description  => 'Coca Cola');
      add_return     (pi_description  => 'Coca Cola');
      add_return     (pi_description  => 'Empty Vodka Bottles'); -- If this variable less than 5 characters => exception

  -- Get a guid to identify errors from this transaction with
  l_transaction_guid := sys_guid;
  
  -- Insert the data into the tables
  insert
  into   ero_clients
         (id  
         ,name
         )
  select id
  ,      name
  from   table (l_clients_nt)
  log errors (l_transaction_guid) reject limit unlimited
  ;

  handle_bulk_errors (pi_table_name => c_tbl_clients);

  insert
  into   ero_subscriptions
         (id          
         ,client_id   
         ,description 
         )
  select id          
  ,      client_id   
  ,      description 
  from   table (l_subscriptions_nt)
  log errors (l_transaction_guid) reject limit unlimited
  ;
  
  handle_bulk_errors (pi_table_name => c_tbl_subscriptions);

  insert
  into   ero_deliveries
         (id                
         ,subscription_id   
         ,description       
         )
  select id                
  ,      subscription_id   
  ,      description       
  from   table (l_deliveries_nt)
  log errors (l_transaction_guid) reject limit unlimited
  ;

  handle_bulk_errors (pi_table_name => c_tbl_deliveries);

  insert
  into   ero_returns
         (id                
         ,subscription_id   
         ,description       
         )
  select id                
  ,      subscription_id   
  ,      description       
  from   table (l_returns_nt)
  log errors (l_transaction_guid) reject limit unlimited
  ;

  handle_bulk_errors (pi_table_name => c_tbl_returns);

  commit;

end;
/

------------------------------------------------------------------
-- Query the tables
------------------------------------------------------------------
select * from ero_returns       order by id;
select * from ero_deliveries    order by id;
select * from ero_subscriptions order by id;
select * from ero_clients       order by id;


------------------------------------------------------------------
-- Cleanup the tables before a new run
------------------------------------------------------------------
delete from ero_returns           ;
delete from ero_deliveries        ;
delete from ero_subscriptions     ;
delete from ero_clients           ;
delete from err$_ero_returns      ;
delete from err$_ero_deliveries   ;
delete from err$_ero_subscriptions;
delete from err$_ero_clients      ;
commit;
