------------------------------------------------------------------
-- drop objects if they exist
------------------------------------------------------------------
drop table ero_deliveries;
drop table ero_returns;
drop table ero_subscriptions;
drop table ero_clients;
drop type ero_mapping_ntt;
drop type ero_mapping_ot;

------------------------------------------------------------------
-- create tables, PK's and UK's
------------------------------------------------------------------
create table ero_clients
  (id      number(10,0)   not null
  ,name    varchar2(30)   not null
  );

create unique index ero_clients_pk on ero_clients (id);
alter table ero_clients add (constraint ero_clients_pk primary key (id) using index);

create table ero_subscriptions
  (id          number(10,0)   not null
  ,client_id   number(10,0)   not null
  ,description varchar2(30)   not null
  );

create unique index ero_subscriptions_pk on ero_subscriptions (id);
alter table ero_subscriptions add constraint ero_subscriptions_pk   primary key (id) using index;
alter table ero_subscriptions add constraint ero_subscr_clients_fk1 foreign key (client_id) references ero_clients (id);

create table ero_deliveries
  (id                number(10,0)   not null
  ,subscription_id   number(10,0)   not null
  ,description       varchar2(30)   not null
  );

create unique index ero_deliveries_pk on ero_deliveries (id);
alter table ero_deliveries add constraint ero_deliveries_pk         primary key (id) using index;
alter table ero_deliveries add constraint ero_deliveries_subscr_fk1 foreign key (subscription_id) references ero_subscriptions (id);

create table ero_returns
  (id                number(10,0)   not null
  ,subscription_id   number(10,0)   not null
  ,description       varchar2(30)   not null
  );

create unique index ero_returns_pk on ero_returns (id);
alter table ero_returns add constraint ero_returns_pk         primary key (id) using index;
alter table ero_returns add constraint ero_returns_subscr_fk1 foreign key (subscription_id) references ero_subscriptions (id);

------------------------------------------------------------------
-- Create a check constraint to generate an error in the forall
------------------------------------------------------------------
alter table ero_returns
add (constraint ero_returns_ck1
     check (-- description must be at least 5 chracters
            length(description) >= 5
           )
    );

------------------------------------------------------------------
-- Create nested table for mapping
------------------------------------------------------------------
create or replace type ero_mapping_ot force as object
  (table_name         varchar2(30)
  ,collection_index   integer
  ,top_level_pk       number(10,0)
  ,table_pk           number(10,0)
  );
/

create or replace type ero_mapping_ntt force as table of ero_mapping_ot
/

------------------------------------------------------------------
-- Execute the Multi-Forall code
------------------------------------------------------------------
declare
  subtype st_identifier   is varchar2(30);
  
  c_tbl_clients         constant   st_identifier    := 'ero_clients'      ;
  c_tbl_subscriptions   constant   st_identifier    := 'ero_subscriptions';
  c_tbl_deliveries      constant   st_identifier    := 'ero_deliveries'   ;
  c_tbl_returns         constant   st_identifier    := 'ero_returns'      ;
  
  type l_clients_aat       is table of ero_clients%rowtype       index by pls_integer;
  type l_subscriptions_aat is table of ero_subscriptions%rowtype index by pls_integer;
  type l_deliveries_aat    is table of ero_deliveries%rowtype    index by pls_integer;
  type l_returns_aat       is table of ero_returns%rowtype       index by pls_integer;
  
  l_clients_aa         l_clients_aat       ;
  l_subscriptions_aa   l_subscriptions_aat ;
  l_deliveries_aa      l_deliveries_aat    ;
  l_returns_aa         l_returns_aat       ;
  
  l_client_id          ero_clients.id%type        := 0;
  l_subscription_id    ero_subscriptions.id%type  := 0;
  l_delivery_id        ero_deliveries.id%type     := 0;
  l_return_id          ero_returns.id%type        := 0;
  
  l_mapping_nt         ero_mapping_ntt   := ero_mapping_ntt();

  e_bulk_errors    exception;
  pragma exception_init(e_bulk_errors, -24381);
  
  
  -- Print a line
  procedure pl 
    (pi_line  in  varchar2  default ''
    )
    is
    begin   -- pl
      dbms_output.put_line (pi_line);
    end pl;
  

  -- Add a record for a collection record to the mapping nested table
  procedure add_to_mapping
    (pi_table_name        in  varchar2
    ,pi_collection_index  in  integer     
    ,pi_table_pk          in  number
    )
    is
    begin   -- add_to_mapping
      l_mapping_nt.extend;
      l_mapping_nt(l_mapping_nt.last) := ero_mapping_ot(table_name       => pi_table_name      
                                                       ,collection_index => pi_collection_index
                                                       ,top_level_pk     => l_client_id
                                                       ,table_pk         => pi_table_pk        
                                                       );
    end add_to_mapping;
  
  
  -- Add a client to its associative array
  procedure add_client
    (pi_name  in  ero_clients.name%type
    )
    is
      l_client_rec  ero_clients%rowtype;
    begin   -- add_client
      -- increase id
      l_client_id := l_client_id + 1;
      
      -- Add the client to the associative array
      l_client_rec.id   := l_client_id;
      l_client_rec.name := pi_name    ;
      l_clients_aa(l_clients_aa.count + 1) := l_client_rec;
      
      -- Add a record that maps this associative array row to the top level PK
      add_to_mapping (pi_table_name       => c_tbl_clients
                     ,pi_collection_index => l_clients_aa.last
                     ,pi_table_pk         => l_client_id
                     );
    end add_client;
  

  -- Add a subscription to its associative array
  procedure add_subscription
    (pi_description  in  ero_subscriptions.description%type
    )
    is
      l_subscription_rec  ero_subscriptions%rowtype;
    begin   -- add_subscription
      -- increase id
      l_subscription_id := l_subscription_id + 1;
      
      -- Add the subscription to the associative array
      l_subscription_rec.id          := l_subscription_id;
      l_subscription_rec.client_id   := l_client_id      ;
      l_subscription_rec.description := pi_description   ;
      l_subscriptions_aa(l_subscriptions_aa.count + 1) := l_subscription_rec;
      
      -- Add a record that maps this associative array row to the top level PK
      add_to_mapping (pi_table_name       => c_tbl_subscriptions
                     ,pi_collection_index => l_subscriptions_aa.last
                     ,pi_table_pk         => l_subscription_id
                     );
    end add_subscription;
  

  -- Add a delivery to its associative array
  procedure add_delivery
    (pi_description      in  ero_deliveries.description%type
    )
    is
      l_delivery_rec  ero_deliveries%rowtype;
    begin   -- add_delivery
      -- increase id
      l_delivery_id := l_delivery_id + 1;
      
      -- Add the delivery to the associative array
      l_delivery_rec.id                := l_delivery_id    ;
      l_delivery_rec.subscription_id   := l_subscription_id;
      l_delivery_rec.description       := pi_description   ;
      l_deliveries_aa(l_deliveries_aa.count + 1) := l_delivery_rec;
      
      -- Add a record that maps this associative array row to the top level PK
      add_to_mapping (pi_table_name       => c_tbl_deliveries
                     ,pi_collection_index => l_deliveries_aa.last
                     ,pi_table_pk         => l_delivery_id
                     );
    end add_delivery;
  

  -- Add a return to its associative array
  procedure add_return
    (pi_description      in  ero_returns.description%type
    )
    is
      l_return_rec  ero_returns%rowtype;
    begin   -- add_return
      -- increase id
      l_return_id := l_return_id + 1;
      
      -- Add the return to the associative array
      l_return_rec.id                := l_return_id      ;
      l_return_rec.subscription_id   := l_subscription_id;
      l_return_rec.description       := pi_description   ;
      l_returns_aa(l_returns_aa.count + 1) := l_return_rec;
      
      -- Add a record that maps this associative array row to the top level PK
      add_to_mapping (pi_table_name       => c_tbl_returns
                     ,pi_collection_index => l_returns_aa.last
                     ,pi_table_pk         => l_return_id
                     );
    end add_return;
  

  procedure handle_bulk_errors
    (pi_table_name  in  st_identifier
    )
    is
      l_error_index    integer;
      l_error_code     integer;
      l_table_pk       number(10,0);
    begin   -- handle_bulk_errors
      for i_err in 1 .. sql%bulk_exceptions.count
      loop
        pl ('ERROR occured while processing table: '||pi_table_name);
        pl ('=====================================================================');
        
        l_error_index := sql%bulk_exceptions(i_err).error_index;
        l_error_code  := sql%bulk_exceptions(i_err).error_code ;
        
        pl ('Error: '||sqlerrm(-1 * l_error_code));
        
        -- Get the PK for the record that caused the exception
        l_table_pk    := case pi_table_name
                           when c_tbl_clients       then l_clients_aa      (l_error_index).id
                           when c_tbl_subscriptions then l_subscriptions_aa(l_error_index).id
                           when c_tbl_deliveries    then l_deliveries_aa   (l_error_index).id
                           when c_tbl_returns       then l_returns_aa      (l_error_index).id
                         end;
        
        -- Delete all work for the Client that may already have been succesfully done
        delete 
        from   ero_returns
        where  id in 
               (select to_remove.table_pk
                from   table (l_mapping_nt)    erroneous
                join   table (l_mapping_nt)    to_remove
                  on   to_remove.top_level_pk  = erroneous.top_level_pk
                where  erroneous.table_name    = pi_table_name
                  and  erroneous.table_pk      = l_table_pk
                  and  to_remove.table_name    = c_tbl_returns
               );
        
        pl ('ero_returns       records removed: '||to_char(sql%rowcount));
        
        delete 
        from   ero_deliveries
        where  id in 
               (select to_remove.table_pk
                from   table (l_mapping_nt)    erroneous
                join   table (l_mapping_nt)    to_remove
                  on   to_remove.top_level_pk  = erroneous.top_level_pk
                where  erroneous.table_name    = pi_table_name
                  and  erroneous.table_pk      = l_table_pk
                  and  to_remove.table_name    = c_tbl_deliveries
               );
        
        pl ('ero_deliveries    records removed: '||to_char(sql%rowcount));
        
        delete 
        from   ero_subscriptions
        where  id in 
               (select to_remove.table_pk
                from   table (l_mapping_nt)    erroneous
                join   table (l_mapping_nt)    to_remove
                  on   to_remove.top_level_pk  = erroneous.top_level_pk
                where  erroneous.table_name    = pi_table_name
                  and  erroneous.table_pk      = l_table_pk
                  and  to_remove.table_name    = c_tbl_subscriptions
               );
        
        pl ('ero_subscriptions records removed: '||to_char(sql%rowcount));
        
        delete 
        from   ero_clients
        where  id in 
               (select to_remove.table_pk
                from   table (l_mapping_nt)    erroneous
                join   table (l_mapping_nt)    to_remove
                  on   to_remove.top_level_pk  = erroneous.top_level_pk
                where  erroneous.table_name    = pi_table_name
                  and  erroneous.table_pk      = l_table_pk
                  and  to_remove.table_name    = c_tbl_clients
               );
        
        pl ('ero_clients       records removed: '||to_char(sql%rowcount));
        
      end loop;
    
      -- Don't forget to remove records from collections that may not have been 
      -- processed yet
      for r_del
      in  (select to_remove.table_name
           ,      to_remove.collection_index
           from   table (l_mapping_nt)    erroneous
           join   table (l_mapping_nt)    to_remove
             on   to_remove.top_level_pk  = erroneous.top_level_pk
           where  erroneous.table_name    = pi_table_name
             and  erroneous.table_pk      = l_table_pk
             and  to_remove.table_name   in (c_tbl_clients
                                            ,c_tbl_subscriptions
                                            ,c_tbl_deliveries
                                            ,c_tbl_returns
                                            )
          )
      loop
        case r_del.table_name
          when c_tbl_clients       then l_clients_aa      .delete (r_del.collection_index);
                                        pl ('1 Row removed from Clients       collection');
          when c_tbl_subscriptions then l_subscriptions_aa.delete (r_del.collection_index);
                                        pl ('1 Row removed from Subscriptions collection');
          when c_tbl_deliveries    then l_deliveries_aa   .delete (r_del.collection_index);
                                        pl ('1 Row removed from Deliveries    collection');
          when c_tbl_returns       then l_returns_aa      .delete (r_del.collection_index);
                                        pl ('1 Row removed from Returns       collection');
        end case;
      end loop;
    
    end handle_bulk_errors;
  
begin
  -- prepare data for 3 clients
  -- one returns-record for client 2 will raise an exception
  add_client         (pi_name         => 'My very first client');
    add_subscription (pi_description  => 'Fruit');
      add_delivery   (pi_description  => 'Appels');
      add_delivery   (pi_description  => 'Pears');
      add_return     (pi_description  => 'Grapes');
    
    add_subscription (pi_description  => 'Beverages');
      add_delivery   (pi_description  => 'Beer');
      add_delivery   (pi_description  => 'Water');
      add_return     (pi_description  => 'Water');
  
  add_client         (pi_name         => 'Problematic Client');
    add_subscription (pi_description  => 'Fruit');
      add_delivery   (pi_description  => 'Strawberries');
      add_delivery   (pi_description  => 'Cherries');
      add_return     (pi_description  => 'Strawberries');
  
    add_subscription (pi_description  => 'Beverages');
      add_delivery   (pi_description  => 'whiskey');
      add_delivery   (pi_description  => 'Lemonade');
      add_return     (pi_description  => 'Lemonade');
      add_return     (pi_description  => '&Returned_Item'); -- If this variable less than 5 characters => exception
  
  add_client         (pi_name         => 'Latest Client');
    add_subscription (pi_description  => 'Fruit');
      add_delivery   (pi_description  => 'Oranges');
      add_delivery   (pi_description  => 'Bananas');
      add_return     (pi_description  => 'Bananas');
  
    add_subscription (pi_description  => 'Beverages');
      add_delivery   (pi_description  => 'Vodka');
      add_delivery   (pi_description  => 'Coca Cola');
      add_return     (pi_description  => 'Coca Cola');
      add_return     (pi_description  => 'Empty Vodka Bottles');
  
  -- Use Forall to write the data to the tables
  begin
    forall i_cli in indices of l_clients_aa save exceptions
      insert
      into   ero_clients
      values l_clients_aa(i_cli)
    ;
  exception
    when e_bulk_errors
    then
      handle_bulk_errors (pi_table_name => c_tbl_clients);
  end;

  begin
    forall i_sub in indices of l_subscriptions_aa save exceptions
      insert
      into   ero_subscriptions
      values l_subscriptions_aa(i_sub)
    ;
  exception
    when e_bulk_errors
    then
      handle_bulk_errors (pi_table_name => c_tbl_subscriptions);
  end;

  begin
    forall i_del in indices of l_deliveries_aa save exceptions
      insert
      into   ero_deliveries
      values l_deliveries_aa(i_del)
    ;
  exception
    when e_bulk_errors
    then
      handle_bulk_errors (pi_table_name => c_tbl_deliveries);
  end;

  begin
    forall i_ret in indices of l_returns_aa save exceptions
      insert
      into   ero_returns
      values l_returns_aa(i_ret)
    ;
  exception
    when e_bulk_errors
    then
      handle_bulk_errors (pi_table_name => c_tbl_returns);
  end;

  -- NOTE: This is only a demo of the multiple forall functionality
  --       It has just 1 error, so all is well.
  --       But after handling the first error the collections are sparse
  --       So real code needs to handle that similar to earlier in the presentation

  commit;

end;
/

------------------------------------------------------------------
-- Query the tables
------------------------------------------------------------------
select * from ero_returns       order by id;
select * from ero_deliveries    order by id;
select * from ero_subscriptions order by id;
select * from ero_clients       order by id;


------------------------------------------------------------------
-- Cleanup the tables before a new run
------------------------------------------------------------------
delete from ero_returns      ;
delete from ero_deliveries   ;
delete from ero_subscriptions;
delete from ero_clients      ;
commit;
